using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Collections;

namespace ConvexHulls
{
    /// <summary>
    /// Convex hull algorithm "Mergehull".
    /// </summary>
    class MergeHull : IConvexHull
    {
        private class DoubleList
        {
            public class ListElem
            {
                public Point Point = Point.Empty;
                public ListElem Previous = null;
                public ListElem Next = null;

                public ListElem(Point p)
                {
                    Point = p;
                }

                public ListElem(Point p, ListElem prev)
                    : this(p)
                {
                    Previous = prev;
                }

                public ListElem(Point p, ListElem prev, ListElem next)
                    : this(p, prev)
                {
                    Next = next;
                }
            }


            private ListElem first = null;
            private ListElem last = null;
            public ListElem Current = null;
            private int count = 0;

            public ListElem First
            {
                get { return first; }
            }

            public ListElem Last
            {
                get { return last; }
            }

            public int Count
            {
                get { return count; }
            }


            public void Add(Point p)
            {
                ListElem elem = new ListElem(p);
                this.Add(elem);
            }

            public void Add(ListElem elem)
            {
                elem.Previous = last;
                elem.Next = null;
                if (first == null)
                    first = elem;
                else
                    last.Next = elem;
                last = elem;
                count++;
            }

            public void AddRange(DoubleList list)
            {
                ListElem cur = list.first;
                while (cur != null)
                {
                    this.Add(cur);
                    cur = cur.Next;
                }
            }
        }

        /// <summary>
        /// Compares two points with their x coordinates.
        /// </summary>
        private class XComparer : IComparer<Point>
        {
            public int Compare(Point p1, Point p2)
            {
                int res = p1.X.CompareTo(p2.X);
                return (res == 0) ? p1.Y.CompareTo(p2.Y) : res;
            }
        }

        /// <summary>
        /// Calculates the convex hull.
        /// </summary>
        public Point[] CalculateHull(Point[] points)
        {
            // trivial cases
            if (points.Length <= 2)
                return points;

            // create an array copy
            Point[] points2 = new Point[points.Length];
            for (int i = 0; i < points.Length; i++)
                points2[i] = points[i];
            // sort the array copy
            Array.Sort(points2, new XComparer());

            // recursive calculation of the hull
            DoubleList listhull = mhullRec(points2, 0, points2.Length - 1);

            // create result array
            Point[] hull = new Point[listhull.Count];
            int ii = 0;
            DoubleList.ListElem elem = listhull.First;
            while (elem != null)
            {
                hull[ii++] = elem.Point;
                elem = elem.Next;
            }

            return hull;
        }

        private DoubleList mhullRec(Point[] points, int l, int r)
        {
            // trivial cases
            if (r-l+1 <= 2)
            {
                DoubleList list = new DoubleList();
                for (int i = l; i <= r; i++)
                    list.Add(points[i]);
                return list;
            }

            // divide and conquer
            int m = (r + l) / 2;
            DoubleList lhull = mhullRec(points, l, m);
            DoubleList rhull = mhullRec(points, m + 1, r);

            // combine
            // search in left list for max
            DoubleList.ListElem curl = lhull.First.Next;
            DoubleList.ListElem maxl = lhull.First;
            while (curl != null)
            {
                if (curl.Point.X > maxl.Point.X || (curl.Point.X == maxl.Point.X && curl.Point.Y < maxl.Point.Y))
                    maxl = curl;
                curl = curl.Next;
            }
            // search in right list for min
            DoubleList.ListElem curr = rhull.First.Next;
            DoubleList.ListElem minr = rhull.First;
            while (curr != null)
            {
                if (curr.Point.X < minr.Point.X || (curr.Point.X == minr.Point.X && curr.Point.Y > minr.Point.Y))
                    minr = curr;
                curr = curr.Next;
            }
            // find lower tangent
            curl = maxl;
            curr = minr;
            bool end = false;
            while (!end)
            {
                end = true;
                if (!isLowerTangent(curl, curr, lhull, true))
                {
                    if (curl == lhull.First)
                        curl = lhull.Last;
                    else
                        curl = curl.Previous;
                    end = false;
                }
                if (!isLowerTangent(curl, curr, rhull, false))
                {
                    if (curr == rhull.Last)
                        curr = rhull.First;
                    else
                        curr = curr.Next;
                    end = false;
                }
            }
            DoubleList.ListElem llower = curl;
            DoubleList.ListElem rlower = curr;
            // find upper tangent
            curl = maxl;
            curr = minr;
            end = false;
            while (!end)
            {
                end = true;
                if (!isUpperTangent(curl, curr, lhull, true))
                {
                    if (curl == lhull.Last)
                        curl = lhull.First;
                    else
                        curl = curl.Next;
                    end = false;
                }
                if (!isUpperTangent(curl, curr, rhull, false))
                {
                    if (curr == rhull.First)
                        curr = rhull.Last;
                    else
                        curr = curr.Previous;
                    end = false;
                }
            }
            DoubleList.ListElem lupper = curl;
            DoubleList.ListElem rupper = curr;
            // combine to a new hull
            DoubleList hull = new DoubleList();
            // elements in lhull from lupper to llower
            DoubleList.ListElem cur = lupper;
            hull.Add(new DoubleList.ListElem(lupper.Point));
            while(cur != llower)
            {
                if (cur == lhull.Last)
                    cur = lhull.First;
                else
                    cur = cur.Next;
                hull.Add(new DoubleList.ListElem(cur.Point));
            }
            // elements in rhull from rlower to rupper
            cur = rlower;
            hull.Add(new DoubleList.ListElem(rlower.Point));
            while(cur != rupper)
            {
                if (cur == rhull.Last)
                    cur = rhull.First;
                else
                    cur = cur.Next;
                hull.Add(new DoubleList.ListElem(cur.Point));
            }

            return hull;
        }

        private bool isLowerTangent(DoubleList.ListElem l, DoubleList.ListElem r, DoubleList hull, bool left)
        {
            DoubleList.ListElem prev = left ? l.Previous : r.Previous;
            DoubleList.ListElem next = left ? l.Next : r.Next;
            if (prev == null)
                prev = hull.Last;
            if (next == null)
                next = hull.First;

            int distprev = getDist(prev.Point, l.Point, r.Point);
            int distnext = getDist(next.Point, l.Point, r.Point);

            return (distprev <= 0 && distnext <= 0);
            //return (distprev <= 0 && (distnext < 0 || distnext == 0 && (left ? !isLeftOf(next.Point, l.Point, r.Point) : !isRightOf(next.Point, l.Point, r.Point))));
            //return (!isPLeft(prev.Point, r.Point, l.Point) && !isPLeft(next.Point, r.Point, l.Point));
        }

        private bool isUpperTangent(DoubleList.ListElem l, DoubleList.ListElem r, DoubleList hull, bool left)
        {
            return isLowerTangent(r, l, hull, !left);
        }

        /*private bool isLeftOf(Point p, Point vfrom, Point vto)
        {
            int dx = vto.X - vfrom.X;
            int dy = vto.Y - vfrom.Y;
            return (dx == 0) ? (dy > 0) ? p.Y < vfrom.Y : p.Y > vfrom.Y : (dx < 0) ? p.X - vto.X < vfrom.X - vto.X : vto.X - p.X > dx;
        }

        private bool isRightOf(Point p, Point vfrom, Point vto)
        {
            int dx = vto.X - vfrom.X;
            int dy = vto.Y - vfrom.Y;
            return (dx == 0) ? (dy > 0) ? p.Y > vto.Y : p.Y < vto.Y : (dx < 0) ? p.X - vfrom.X < vto.X - vfrom.X : p.X - vfrom.X > dx;
        }*/

        /// <summary>
        /// Returns the distance of a point p to a vector (from,to).
        /// </summary>
        private int getDist(Point p, Point vfrom, Point vto)
        {
            return ((vto.X - vfrom.X) * (p.Y - vfrom.Y) - (p.X - vfrom.X) * (vto.Y - vfrom.Y));
        }

        /// <summary>
        /// Returns true, if the point p is on the left hand side of the vector vfrom->vto.
        /// </summary>
        private bool isPLeft(Point p, Point vfrom, Point vto)
        {
            return (getDist(p, vfrom, vto) < 0);
        }
    }
}
